#version 330

uniform sampler2D gb_diffuse;
uniform sampler2D gb_normal;
uniform sampler2D gb_depth;
uniform sampler2D lightPass_diffuse;
uniform sampler2D lightPass_specular;
uniform mat4 viewWorldTrans; // inverse world-view-projection transform
uniform int lightType;
uniform vec4 lightPosition;
uniform float lightRadius;
uniform vec3 lightDirection;
uniform vec3 lightColor;
uniform vec3 lightAttenuation;
uniform vec3 camPos;

// shadow mapping uniforms
uniform samplerCubeShadow shadowDepthCm;
//uniform samplerCube shadowDepthCm;
uniform vec3 shadowNearFar;
uniform int hasShadow;

in vec2 fragmentUV;
in vec2 fragmentPos;

out vec4 diffuse_out;
out vec4 specular_out;

const float k_a = 0.01; // ambient coefficient
const float k_d = 0.8; // diffuse coefficient
const float k_s = 0.7; // specular reflection coefficient
const float phongExp = 3; // specular parameter

const float shadowEps = 0.02; // shadow depth test eps
const float sr = 0.04;
const int numShadowSamples = 14;
const vec3 shadowSamples[numShadowSamples] = vec3[numShadowSamples] (
	vec3(0, 0, 0), 
	// normally distributed samples
	vec3(-0.65, 0.08, 1.01), 
	vec3(1.19, -0.95, -2.12), 
	vec3(-1.61, 0.41, -0.50), 
	vec3(-0.02, 0.68, -1.27), 
	vec3(-1.95, 0.86, -0.38), 
	vec3(1.02, -0.69, 0.65), 
	vec3(0.86, 0.45, 0.83), 
	vec3(0.00, 0.10, -1.01), 
	vec3(-0.07, 0.83, -0.47), 
	vec3(-2.49, 0.54, 0.14), 
	vec3(0.58, 0.90, -0.29), 
	vec3(-2.19, -0.13, 0.30), 
	vec3(-2.32, -0.15, 0.40)
);

/*
float calcShadowSimple(vec3 lightToFrag, float lightDist) {
	// standard sampler
	float smallestDistToLight = texture(shadowDepthCm, lightToFrag);
	smallestDistToLight *= shadowNearFar.y;
	if (smallestDistToLight + shadowEps < lightDist) {
		return 0;
	}
	return 1;
}
*/

float calcShadowAutoPCS(vec3 lightToFrag, vec3 lightVector, float lightDist, float farPlaneDist, vec3 normal) {
	float bias = 0.005 * tan(acos(max(dot(lightVector, normal), 0))); 
	bias = clamp(bias, 0, 0.005);
	return texture(shadowDepthCm, vec4(lightToFrag, lightDist / farPlaneDist - bias));

	/*
	float shadowFactor = texture(shadowDepthCm, vec4(lightToFrag, lightDist / farPlaneDist - bias));
	for (int i = 0; i < numShadowSamples; ++i) {
		shadowFactor += texture(shadowDepthCm, vec4(lightToFrag + sr * shadowSamples[i], length(lightToFrag + sr * shadowSamples[i]) / farPlaneDist - bias));
	}
	shadowFactor /= numShadowSamples;
	return shadowFactor;
	*/
}

/*
float calcShadowPCS(vec3 lightToFrag, float lightDist) {
	float smallestDistToLight = texture(shadowDepthCm, lightToFrag);
	float shadowFactor = 0;
	for (int i = 0; i < numShadowSamples; ++i) {
		smallestDistToLight = texture(shadowDepthCm, lightToFrag + sr * shadowSamples[i]).r;
		if (smallestDistToLight * shadowNearFar.y > lightDist - shadowEps) {
			shadowFactor += 1;
		}
	}
	shadowFactor /= numShadowSamples;

	return shadowFactor;
}
*/

/*
float calcShadowVSM(vec3 lightToFrag, float lightDist) {
	vec2 moments = 0.25 * texture(shadowDepthCm, lightToFrag).rg;
	for (int i = 0; i < 12; ++i) {
		moments += 0.75 / 12 * texture(shadowDepthCm, lightToFrag + sr * shadowSamples[i]).rg;
	}
	float lightDistNorml = lightDist / shadowNearFar.y;
	
	// fragment is visible
	if (lightDistNorml < moments.x) {
		return 1;
	}

	// chebychev
	float variance = moments.y - (moments.x * moments.x);
	variance = max(variance, 0.000002);
	float d = lightDistNorml - moments.x;
	float p_max = variance / (variance + d * d);
	return max(smoothstep(0.6, 1, p_max), 0.05);
}
*/

void main() {
	vec4 diffuseColor = texture(gb_diffuse, fragmentUV);
	vec3 normal = texture(gb_normal, fragmentUV).rgb * 2 - 1;
	float depth = texture(gb_depth, fragmentUV).r;

	vec4 prevPassDiffuse = vec4(texture(lightPass_diffuse, fragmentUV).rgb, 1);
	vec4 prevPassSpecular = vec4(texture(lightPass_specular, fragmentUV).rgb, 1);

	// viewport position of current fragment in [-1,1]
	vec4 H = vec4(fragmentPos.x * 2 - 1, fragmentPos.y * 2 - 1, depth * 2 - 1, 1);
	// world position of current fragment
	vec4 D = viewWorldTrans * H;
	vec4 worldPos = D / D.w;

	// lighting
	vec3 halfWay;
	vec3 lightVector;
	float attFactor = 1;
	float shadowFactor = 1;
	if (lightType == 0) { // point light
		float lightDist = distance(lightPosition.xyz, worldPos.xyz);
		attFactor = 1 / (lightAttenuation.x + lightAttenuation.y * lightDist + lightAttenuation.z * lightDist * lightDist);
		lightVector = normalize(lightPosition.xyz - worldPos.xyz);
		halfWay = normalize(lightVector + normalize(camPos - worldPos.xyz));

		// shadow mapping 
		if (hasShadow == 1 && diffuseColor.a > 0) {
			shadowFactor = calcShadowAutoPCS(worldPos.xyz - lightPosition.xyz, lightVector, lightDist, shadowNearFar.y, normal);
		}
	} else if (lightType == 1) { // directional light
		halfWay = normalize(-lightDirection + camPos);
		lightVector = normalize(-lightDirection);
	}

	float diffuseI = max(dot(lightVector, normal), 0);

	if(diffuseColor.a <= 0) {
		diffuseI = 1.0;
		halfWay = normal;
	}

	float specFactor = diffuseI > 0 ? 1 : 0;
	vec3 diffuseNew = k_a * diffuseColor.rgb + k_d * diffuseI * attFactor * lightColor * diffuseColor.rgb * shadowFactor;
	vec3 specularNew = specFactor * k_s * pow(max(dot(halfWay, normal), 0), phongExp) * attFactor * lightColor * shadowFactor * diffuseColor.rgb;

	// output
	diffuse_out = vec4(prevPassDiffuse.rgb + diffuseNew, 1);
	specular_out = vec4(prevPassSpecular.rgb + specularNew, 1);
}
